"""Displays the progress of a loop with a command line progress bar"""
from andrewtools.utils import validate_type


def progress_bar(
    iteration: int, total: int, width: int = 100, label: str = "", end: str = ""
) -> None:
    """
    Usage: Call as the last line inside of a loop.

    Note that the progress bar will not play well with your code if there are
    other print statements within the same loop.

    :param int iteration: the current iteration, 0-indexed
    :param int total: the total number of iterations, 1-indexed
    :param int width: the inner width of the progress bar itself [10-100]
    :param str label: label is printed on the left side of the progress bar
    """

    # Validate input types
    iteration = validate_type(iteration, int)
    total = validate_type(total, int)
    width = validate_type(width, int)
    label = validate_type(label, str)
    end = validate_type(end, str)

    # Make sure width is in bounds
    if width < 10 or width > 100:
        raise ValueError("width must be an integer between 10 and 100, inclusive")

    # Make sure iteration and total in bounds
    if iteration < 0 or total < 0:
        raise ValueError("iteration and total must be positive integers")

    # Handle user error allowing iteration to be greater than or equal to total
    if iteration >= total:
        # It sure would suck if this tool to terminated a program that has been
        # running for a long time! Therefore, we will fail silently.
        msg = (
            " > Progress Bar error: iteration > total. "
            + "Note that iteration is 0-indexed and total is 1-indexed. "
            + "The loop will continue..."
        )
        print(msg, end="\r")
        return

    # Calculate percentage complete and create progress bar
    percent = (iteration + 1) / total
    complete = int(width * percent)
    bar = "*" * complete + "-" * (width - complete)
    print(f"\r{label} |{bar}| {percent * 100:.1f}% {end}", end="\r")

    # Final iteration ends with newline character
    if percent == 1:
        print()


if __name__ == "__main__":

    # Usage
    import time

    iterations = 39
    for i in range(iterations):
        time.sleep(0.1)
        progress_bar(i, iterations, width=10, label="Progress")



# shortens import statement from:
#  - `from andrewtools.progress_bar import progress_bar` to
#  - `from andrewtools import progress_bar`
from .progress_bar import progress_bar
from .andrewtimer import AndrewTimer



"""An easy-to-use timer"""
import time
from andrewtools import progress_bar
from andrewtools.utils import validate_type


class AndrewTimer:
    def __init__(self):
        self.start_time = time.time()
        self.lap_start = self.start_time
        self.laps = []

    def elapsed(self, units="s", format=False):
        """
        Return the total elapsed time

        :param str units: 's' = seconds, 'ms' = milliseconds, 'ns' = nanoseconds
                          'm' = minutes
        :param bool format: False = return elapsed time as a float
                            True = return a format string with units appended
        """
        # Validate input types
        units = validate_type(units, str)
        format = validate_type(format, bool)

        # Calculate elapsed time in correct units
        elapsed = time.time() - self.start_time
        elapsed = self._convert(elapsed, units)

        if format:
            return self._format(elapsed, units)
        return elapsed

    def lap(self):
        """End the current lap and start a new lap"""
        now = time.time()
        self.laps.append(now - self.lap_start)
        self.lap_start = now

    def average(self, units="s", format=False):
        """
        Return average lap time

        :param str units: 's' = seconds, 'ms' = milliseconds, 'ns' = nanoseconds
                          'm' = minutes
        :param bool format: False = return elapsed time as a float
                            True = return a format string with units appended
        """
        # Validate input types
        units = validate_type(units, str)
        format = validate_type(format, bool)

        # Calculate average lap time in correct units
        avg = sum(self.laps) / len(self.laps)
        avg = self._convert(avg, units)

        if format:
            return self._format(avg, units)
        return avg

    def get_laps(self):
        return self.laps

    def last_lap(self, units="s", format=False):
        """
        Return the time of the last lap

        :param str units: 's' = seconds, 'ms' = milliseconds, 'ns' = nanoseconds
                          'm' = minutes
        :param bool format: False = return elapsed time as a float
                            True = return a format string with units appended
        """
        # Validate input types
        units = validate_type(units, str)
        format = validate_type(format, bool)

        # Get time of last lap in correct units
        last_lap = self.laps[-1] if self.laps else 0
        last_lap = self._convert(last_lap, units)

        if format:
            return self._format(last_lap, units)
        return last_lap

    def _convert(self, time, units):
        if units == "s":
            return time
        elif units == "ms":
            return time * 1_000
        elif units == "ns":
            return time * 1_000_000
        elif units == "m":
            return time / 60
        else:
            raise ValueError(f"Units must be 'm', 's', 'ms', or 'ns', not {units}")

    def _format(self, time, units):
        return f"{time:,.3f}{units}"


if __name__ == "__main__":

    # Usage
    t = AndrewTimer()
    for i in range(10):
        time.sleep(0.5)
        t.lap()
        end = f"(Last {t.last_lap(format=True)}) (Total: {t.elapsed(format=True)})"
        progress_bar(i, 10, width=10, label="Progress", end=end)

    total_time = t.elapsed("s", format=True)
    average_time = t.average("s", format=True)
    print(f"Total: {total_time}, Average: {average_time}")  # approx. 5.000s 0.500s



"""Collection of functions used to validate function arguments"""


def validate_type(param_val: any, expected_type: type) -> int:
    """
    Check type of value by attempting to convert to expected type.
    Return converted type, if successful.
    """
    try:
        param_val = expected_type(param_val)
        return param_val
    except (ValueError, TypeError):
        raise
